from __future__ import annotations

import logging
from inspect import signature
from pathlib import Path

import h5py

from .... import compression, types
from ... import datatype, utils
from ...exceptions import LH5EncodeError
from .array import _h5_write_array
from .scalar import _h5_write_scalar
from .vector_of_vectors import _h5_write_vector_of_vectors

log = logging.getLogger(__name__)


def _h5_write_lgdo(
    obj,
    name,
    lh5_file,
    group="/",
    start_row=0,
    n_rows=None,
    wo_mode="append",
    write_start=0,
    **h5py_kwargs,
):
    assert isinstance(obj, types.LGDO)

    file_kwargs = {
        k: h5py_kwargs[k] for k in h5py_kwargs & signature(h5py.File).parameters.keys()
    }
    h5py_kwargs = {k: h5py_kwargs[k] for k in h5py_kwargs - file_kwargs.keys()}
    if wo_mode == "write_safe":
        wo_mode = "w"
    if wo_mode == "append":
        wo_mode = "a"
    if wo_mode == "overwrite":
        wo_mode = "o"
    if wo_mode == "overwrite_file":
        wo_mode = "of"
        write_start = 0
    if wo_mode == "append_column":
        wo_mode = "ac"
    if wo_mode not in ["w", "a", "o", "of", "ac"]:
        msg = f"unknown wo_mode '{wo_mode}'"
        raise LH5EncodeError(msg, lh5_file, group, name)

    # "mode" is for the h5df.File and wo_mode is for this function
    # In hdf5, 'a' is really "modify" -- in addition to appending, you can
    # change any object in the file. So we use file:append for
    # write_object:overwrite.
    if not isinstance(lh5_file, h5py.File):
        mode = "w" if wo_mode == "of" or not Path(lh5_file).exists() else "a"
        lh5_file = h5py.File(lh5_file, mode=mode, **file_kwargs)

    log.debug(
        f"writing {obj!r}[{start_row}:{n_rows}] as "
        f"{lh5_file.filename}:{group}/{name}[{write_start}:], "
        f"mode = {wo_mode}, h5py_kwargs = {h5py_kwargs}"
    )

    group = utils.get_h5_group(group, lh5_file)

    if wo_mode == "w" and name in group:
        msg = f"can't overwrite '{name}' in wo_mode 'write_safe'"
        raise LH5EncodeError(msg, lh5_file, group, name)

    # struct, table, waveform table or histogram.
    if isinstance(obj, types.Struct):
        if (
            isinstance(obj, types.Histogram)
            and wo_mode not in ["w", "o", "of"]
            and name in group
        ):
            msg = f"can't append-write to histogram in wo_mode '{wo_mode}'"
            raise LH5EncodeError(msg, lh5_file, group, name)
        if isinstance(obj, types.Histogram) and write_start != 0:
            msg = f"can't write histogram in wo_mode '{wo_mode}' with write_start != 0"
            raise LH5EncodeError(msg, lh5_file, group, name)

        return _h5_write_struct(
            obj,
            name,
            lh5_file,
            group=group,
            start_row=start_row,
            n_rows=n_rows,
            wo_mode=wo_mode,
            write_start=write_start,
            **h5py_kwargs,
        )

    # scalars
    if isinstance(obj, types.Scalar):
        return _h5_write_scalar(obj, name, lh5_file, group, wo_mode)

    # vector of encoded vectors
    if isinstance(
        obj, (types.VectorOfEncodedVectors, types.ArrayOfEncodedEqualSizedArrays)
    ):
        group = utils.get_h5_group(
            name, group, grp_attrs=obj.attrs, overwrite=(wo_mode == "o")
        )

        # ask not to further compress flattened_data, it is already compressed!
        obj.encoded_data.flattened_data.attrs["compression"] = None

        _h5_write_vector_of_vectors(
            obj.encoded_data,
            "encoded_data",
            lh5_file,
            group=group,
            start_row=start_row,
            n_rows=n_rows,
            wo_mode=wo_mode,
            write_start=write_start,
            **h5py_kwargs,
        )

        if isinstance(obj.decoded_size, types.Scalar):
            _h5_write_scalar(
                obj.decoded_size,
                "decoded_size",
                lh5_file,
                group=group,
                wo_mode=wo_mode,
            )
        else:
            _h5_write_array(
                obj.decoded_size,
                "decoded_size",
                lh5_file,
                group=group,
                start_row=start_row,
                n_rows=n_rows,
                wo_mode=wo_mode,
                write_start=write_start,
                **h5py_kwargs,
            )

        return None

    # vector of vectors
    if isinstance(obj, types.VectorOfVectors):
        return _h5_write_vector_of_vectors(
            obj,
            name,
            lh5_file,
            group=group,
            start_row=start_row,
            n_rows=n_rows,
            wo_mode=wo_mode,
            write_start=write_start,
            **h5py_kwargs,
        )

    # if we get this far, must be one of the Array types
    if isinstance(obj, types.Array):
        return _h5_write_array(
            obj,
            name,
            lh5_file,
            group=group,
            start_row=start_row,
            n_rows=n_rows,
            wo_mode=wo_mode,
            write_start=write_start,
            **h5py_kwargs,
        )

    msg = f"do not know how to write '{name}' of type '{type(obj).__name__}'"
    raise LH5EncodeError(msg, lh5_file, group, name)


def _h5_write_struct(
    obj,
    name,
    lh5_file,
    group="/",
    start_row=0,
    n_rows=None,
    wo_mode="append",
    write_start=0,
    **h5py_kwargs,
):
    assert isinstance(obj, types.Struct)

    # In order to append a column, we need to update the
    # `table{old_fields}` value in `group.attrs['datatype"]` to include
    # the new fields.  One way to do this is to override
    # `obj.attrs["datatype"]` to include old and new fields. Then we
    # can write the fields to the table as normal.
    if wo_mode == "ac":
        old_group = utils.get_h5_group(name, group)
        lgdotype = datatype.datatype(old_group.attrs["datatype"])
        fields = datatype.get_struct_fields(old_group.attrs["datatype"])
        if not issubclass(lgdotype, types.Struct):
            msg = f"Trying to append columns to an object of type {lgdotype.__name__}"
            raise LH5EncodeError(msg, lh5_file, group, name)

        # If the mode is `append_column`, make sure we aren't appending
        # a table that has a column of the same name as in the existing
        # table. Also make sure that the field we are adding has the
        # same size
        if len(list(set(fields).intersection(set(obj.keys())))) != 0:
            msg = (
                f"Can't append {list(set(fields).intersection(set(obj.keys())))} "
                "column(s) to a table with the same field(s)"
            )
            raise LH5EncodeError(msg, lh5_file, group, name)
        # It doesn't matter what key we access, as all fields in the old table have the same size
        if old_group[next(iter(old_group.keys()))].size != obj.size:
            msg = (
                f"Table sizes don't match. Trying to append column of size {obj.size} "
                f"to a table of size {old_group[next(iter(old_group.keys()))].size}."
            )
            raise LH5EncodeError(msg, lh5_file, group, name)

        # Now we can append the obj.keys() to the old fields, and then update obj.attrs.
        fields.extend(list(obj.keys()))
        obj.attrs.pop("datatype")
        obj.attrs["datatype"] = "table" + "{" + ",".join(fields) + "}"

    group = utils.get_h5_group(
        name,
        group,
        grp_attrs=obj.attrs,
        overwrite=(wo_mode in ["o", "ac"]),
    )
    # If the mode is overwrite, then we need to peek into the file's
    # table's existing fields.  If we are writing a new table to the
    # group that does not contain an old field, we should delete that
    # old field from the file
    if wo_mode == "o":
        # Find the old keys in the group that are not present in the
        # new table's keys, then delete them
        for key in list(set(group.keys()) - set(obj.keys())):
            log.debug(f"{key} is not present in new table, deleting field")
            del group[key]

    for field in obj:
        # eventually compress waveform table values with LGDO's
        # custom codecs before writing
        # if waveformtable.values.attrs["compression"] is NOT a
        # WaveformCodec, just leave it there
        obj_fld = None
        if (
            isinstance(obj, types.WaveformTable)
            and field == "values"
            and not isinstance(obj.values, types.VectorOfEncodedVectors)
            and not isinstance(obj.values, types.ArrayOfEncodedEqualSizedArrays)
            and "compression" in obj.values.attrs
            and isinstance(obj.values.attrs["compression"], compression.WaveformCodec)
        ):
            codec = obj.values.attrs["compression"]
            obj_fld = compression.encode(obj.values, codec=codec)
        else:
            obj_fld = obj[field]

        # Convert keys to string for dataset names
        f = str(field)
        _h5_write_lgdo(
            obj_fld,
            f,
            lh5_file,
            group=group,
            start_row=start_row,
            n_rows=n_rows,
            wo_mode=wo_mode,
            write_start=write_start,
            **h5py_kwargs,
        )



from __future__ import annotations

import logging

from .... import types
from ...exceptions import LH5EncodeError

log = logging.getLogger(__name__)


def _h5_write_scalar(obj, name, lh5_file, group="/", wo_mode="append"):
    assert isinstance(obj, types.Scalar)

    if name in group:
        if wo_mode in ["o", "a"]:
            log.debug(f"overwriting {name} in {group}")
            del group[name]
        else:
            msg = f"tried to overwrite but wo_mode is {wo_mode!r}"
            raise LH5EncodeError(msg, lh5_file, group, name)

    ds = group.create_dataset(name, shape=(), data=obj.value)
    ds.attrs.update(
        {
            k: v.encode("utf-8") if isinstance(v, str) else v
            for k, v in obj.attrs.items()
        }
    )



from __future__ import annotations

import fnmatch
import logging
from copy import copy

import h5py

from . import utils
from .store import LH5Store

log = logging.getLogger(__name__)


def ls(
    lh5_file: str | h5py.Group,
    lh5_group: str = "",
    recursive: bool = False,
) -> list[str]:
    """Return a list of LH5 groups in the input file and group, similar
    to ``ls`` or ``h5ls``. Supports wildcards in group names.


    Parameters
    ----------
    lh5_file
        name of file.
    lh5_group
        group to search. add a ``/`` to the end of the group name if you want to
        list all objects inside that group.
    recursive
        if ``True``, recurse into subgroups.
    """

    log.debug(
        f"Listing objects in '{lh5_file}'"
        + ("" if lh5_group == "" else f" (and group {lh5_group})")
    )

    lh5_st = LH5Store()
    # To use recursively, make lh5_file a h5group instead of a string
    if isinstance(lh5_file, str):
        lh5_file = lh5_st.gimme_file(lh5_file, "r")
        if lh5_group.startswith("/"):
            lh5_group = lh5_group[1:]

    if lh5_group == "":
        lh5_group = "*"

    # get the first group in the group path
    splitpath = lh5_group.split("/", 1)
    # filter out objects that don't match lh5_group pattern
    matchingkeys = fnmatch.filter(lh5_file.keys(), splitpath[0])

    ret = []
    # if there were no "/" in lh5_group just return the result
    if len(splitpath) == 1:
        ret = matchingkeys

    else:
        for key in matchingkeys:
            ret.extend([f"{key}/{path}" for path in ls(lh5_file[key], splitpath[1])])

    if recursive:
        rec_ret = copy(ret)
        for obj in ret:
            try:
                rec_ret += ls(lh5_file, lh5_group=f"{obj}/", recursive=True)
            except AttributeError:
                continue

        return rec_ret

    return ret


def show(
    lh5_file: str | h5py.Group,
    lh5_group: str = "/",
    attrs: bool = False,
    indent: str = "",
    header: bool = True,
    depth: int | None = None,
    detail: bool = False,
) -> None:
    """Print a tree of LH5 file contents with LGDO datatype.

    Parameters
    ----------
    lh5_file
        the LH5 file.
    lh5_group
        print only contents of this HDF5 group.
    attrs
        print the HDF5 attributes too.
    indent
        indent the diagram with this string.
    header
        print `lh5_group` at the top of the diagram.
    depth
        maximum tree depth of groups to print
    detail
        whether to print additional information about how the data is stored

    Examples
    --------
    >>> from lgdo import show
    >>> show("file.lh5", "/geds/raw")
    /geds/raw
    ├── channel · array<1>{real}
    ├── energy · array<1>{real}
    ├── timestamp · array<1>{real}
    ├── waveform · table{t0,dt,values}
    │   ├── dt · array<1>{real}
    │   ├── t0 · array<1>{real}
    │   └── values · array_of_equalsized_arrays<1,1>{real}
    └── wf_std · array<1>{real}
    """
    # check tree depth if we are using it
    if depth is not None and depth <= 0:
        return

    # open file
    if isinstance(lh5_file, str):
        lh5_file = h5py.File(utils.expand_path(lh5_file), "r", locking=False)

    # go to group
    if lh5_group != "/":
        lh5_file = lh5_file[lh5_group]

    if header:
        print(f"\033[1m{lh5_group}\033[0m")  # noqa: T201

    # get an iterator over the keys in the group
    it = iter(lh5_file)
    key = None

    # make sure there is actually something in this file/group
    try:
        key = next(it)  # get first key
    except StopIteration:
        print(f"{indent}└──  empty")  # noqa: T201
        return

    # loop over keys
    while True:
        val = lh5_file[key]
        # we want to print the LGDO datatype
        dtype = val.attrs.get("datatype", default="no datatype")
        if dtype == "no datatype" and isinstance(val, h5py.Group):
            dtype = "HDF5 group"

        _attrs = ""
        if attrs:
            attrs_d = dict(val.attrs)
            attrs_d.pop("datatype", "")
            _attrs = "── " + str(attrs_d) if attrs_d else ""

        # is this the last key?
        killme = False
        try:
            k_new = next(it)  # get next key
        except StopIteration:
            char = "└──"
            killme = True  # we'll have to kill this loop later
        else:
            char = "├──"

        print(f"{indent}{char} \033[1m{key}\033[0m · {dtype} {_attrs}")  # noqa: T201

        if detail and isinstance(val, h5py.Dataset):
            char = "|       "
            if killme:
                char = "        "
            toprint = f"{indent}{char}"
            try:
                toprint += f"\033[3mdtype\033[0m={val.dtype}"
                toprint += f", \033[3mshape\033[0m={val.shape}"
                toprint += f", \033[3mnbytes\033[0m={utils.fmtbytes(val.nbytes)}"
                if (chunkshape := val.chunks) is None:
                    toprint += ", \033[3mnumchunks\033[0m=contiguous"
                else:
                    toprint += f", \033[3mnumchunks\033[0m={val.id.get_num_chunks()}"
                    toprint += f", \033[3mchunkshape\033[0m={chunkshape}"
                toprint += ", \033[3mfilters\033[0m="

                numfilters = val.id.get_create_plist().get_nfilters()
                if numfilters == 0:
                    toprint += "None"
                else:
                    toprint += "("
                    for i in range(numfilters):
                        thisfilter = val.id.get_create_plist().get_filter(i)[3].decode()
                        if "lz4" in thisfilter:
                            thisfilter = "lz4"
                        toprint += f"{thisfilter},"
                    toprint += ")"

            except TypeError:
                toprint += "(scalar)"

            print(toprint)  # noqa: T201

        # if it's a group, call this function recursively
        if isinstance(val, h5py.Group):
            show(
                val,
                indent=indent + ("    " if killme else "│   "),
                header=False,
                attrs=attrs,
                depth=depth - 1 if depth else None,
                detail=detail,
            )

        # break or move to next key
        if killme:
            break

        key = k_new



from __future__ import annotations

import pickle

from lgdo.lh5.exceptions import LH5DecodeError, LH5EncodeError


def test_pickle():
    # test (un-)pickling of LH5 exceptions; e.g. for multiprocessing use.

    ex = LH5EncodeError("message", "file", "group", "name")
    ex = pickle.loads(pickle.dumps(ex))
    assert isinstance(ex, LH5EncodeError)

    ex = LH5DecodeError("message", "file", "obj")
    ex = pickle.loads(pickle.dumps(ex))
    assert isinstance(ex, LH5DecodeError)



from __future__ import annotations

from lgdo import lh5


def test_ls(lgnd_test_data):
    lgnd_file = lgnd_test_data.get_path(
        "lh5/LDQTA_r117_20200110T105115Z_cal_geds_raw.lh5"
    )
    assert lh5.ls(lgnd_file) == ["geds"]
    assert lh5.ls(lgnd_file, "/*/raw") == ["geds/raw"]
    assert lh5.ls(lgnd_file, "geds/raw/") == [
        "geds/raw/baseline",
        "geds/raw/channel",
        "geds/raw/energy",
        "geds/raw/ievt",
        "geds/raw/numtraces",
        "geds/raw/packet_id",
        "geds/raw/timestamp",
        "geds/raw/tracelist",
        "geds/raw/waveform",
        "geds/raw/wf_max",
        "geds/raw/wf_std",
    ]
    assert lh5.ls(lgnd_file, recursive=True) == [
        "geds",
        "geds/raw",
        "geds/raw/baseline",
        "geds/raw/channel",
        "geds/raw/energy",
        "geds/raw/ievt",
        "geds/raw/numtraces",
        "geds/raw/packet_id",
        "geds/raw/timestamp",
        "geds/raw/tracelist",
        "geds/raw/waveform",
        "geds/raw/wf_max",
        "geds/raw/wf_std",
        "geds/raw/tracelist/cumulative_length",
        "geds/raw/tracelist/flattened_data",
        "geds/raw/waveform/dt",
        "geds/raw/waveform/t0",
        "geds/raw/waveform/values",
    ]

    lgnd_file = lgnd_test_data.get_path("lh5/L200-comm-20211130-phy-spms.lh5")
    assert lh5.ls(lgnd_file, "ch5/raw/", recursive=True) == [
        "ch5/raw/abs_delta_mu_usec",
        "ch5/raw/baseline",
        "ch5/raw/channel",
        "ch5/raw/daqenergy",
        "ch5/raw/deadtime",
        "ch5/raw/delta_mu_usec",
        "ch5/raw/dr_maxticks",
        "ch5/raw/dr_start_pps",
        "ch5/raw/dr_start_ticks",
        "ch5/raw/dr_stop_pps",
        "ch5/raw/dr_stop_ticks",
        "ch5/raw/eventnumber",
        "ch5/raw/mu_offset_sec",
        "ch5/raw/mu_offset_usec",
        "ch5/raw/numtraces",
        "ch5/raw/packet_id",
        "ch5/raw/runtime",
        "ch5/raw/timestamp",
        "ch5/raw/to_master_sec",
        "ch5/raw/to_start_sec",
        "ch5/raw/to_start_usec",
        "ch5/raw/tracelist",
        "ch5/raw/ts_maxticks",
        "ch5/raw/ts_pps",
        "ch5/raw/ts_ticks",
        "ch5/raw/waveform",
        "ch5/raw/tracelist/cumulative_length",
        "ch5/raw/tracelist/flattened_data",
        "ch5/raw/waveform/dt",
        "ch5/raw/waveform/t0",
        "ch5/raw/waveform/values",
    ]


def test_show(lgnd_file):
    lh5.show(lgnd_file)
    lh5.show(lgnd_file, "/geds/raw")
    lh5.show(lgnd_file, "geds/raw")
